@@ -1,46 +1,48 @@
+import os
+import gzip
+import pickle
 import random
 import logging
+import pkg_resources
 import numpy as np
 import pandas as pd
 
 from scipy import stats
 from typing import Tuple, Union
 
-logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')
-def col_names(df: pd.DataFrame,
-              cols: list=['ID','Age','Sex']) -> tuple:
-  """Matches required column names with common name variants. 
+def load_model(mdl_path: str) -> Tuple[dict, dict]:
+  with gzip.open(mdl_path, 'rb') as f:
+    return pickle.load(f)
+
+def load_examples(file_name: str=''):
+  """Loads example data and models in the package.
 
   Args:
-    df: pandas dataframe.
-    cols: columns to search for name variants.
+    file_name: either name of the example data saved as .csv or
+      name of the SPARE model saved as .pkl.gz.
 
   Returns:
-    a string or a tuple with matched variants.
+    a tuple containing pandas df and 
   """
-  col_name_variants = {'ID':['ID','id','PTID','participant_id'],
-                       'Age':['Age','age','AGE'],
-                       'Sex':['Sex','sex','SEX']}
-  col_name_variants = {a: col_name_variants[a] for a in cols}
-  for k in col_name_variants.keys():
-    if k not in cols:
-      continue
-    for i in col_name_variants[k]:
-      if i in df.columns:
-        col_name_variants[k] = i
-        break
-  col_not_found = [a for a in col_name_variants.keys() if type(col_name_variants[a])==list]
-  if len(col_not_found) > 0:
-    return logging.error(f'Required columns not found: {col_not_found}')
-  if len(cols) == 1:
-    return col_name_variants[cols[0]]
-  else: 
-    return tuple(col_name_variants.values())
+  logging_basic_config(content_only=True)
+  pkg_path = pkg_resources.resource_filename('spare_scores','')
+  list_data = os.listdir(f'{pkg_path}/data/')
+  list_mdl = os.listdir(f'{pkg_path}/mdl/')
+  if file_name in list_data:
+    return pd.read_csv(f'{pkg_path}/data/{file_name}')
+  elif file_name in list_mdl:
+    return load_model(f'{pkg_path}/mdl/{file_name}')
+  else:
+    logging.info('Available example data:')
+    [logging.info(f' - {a}') for a in list_data]
+    logging.info('Available example SPARE models:')
+    [logging.info(f' - {a}') for a in list_mdl]
 
 def check_train(df: pd.DataFrame, 
                 predictors: list,
                 to_predict: str,
-                pos_group: str = '') -> Tuple[pd.DataFrame, list, str]:
+                pos_group: str = '',
+                verbose: int = 1) -> Tuple[pd.DataFrame, list, str]:
   """Checks training dataframe for errors.
 
   Args:
@@ -52,6 +54,9 @@ def check_train(df: pd.DataFrame,
   Returns:
     a tuple containing 1) filtered dataframe, 2) filtered predictors, 3) SPARE model type.
   """
+  logging_basic_config(verbose)
+  if not {'ID','Age','Sex'}.issubset(set(df.columns)):
+    return logging.error('Please check required columns: ID, Age, Sex.')
   if not set(predictors).issubset(df.columns):
     return logging.error('Not all predictors exist in the input dataframe.')
   if to_predict not in df.columns:
@@ -63,7 +68,6 @@ def check_train(df: pd.DataFrame,
     logging.info('Some participants have invalid predictor variables (i.e. n/a). They will be excluded.')
     df = df.loc[np.sum(pd.isna(df[predictors]), axis=1) == 0].reset_index(drop=True)
 
-  col_id = col_names(df,['ID'])
   if len(df[to_predict].unique()) == 2:
     if pos_group == '':
       return logging.error('"pos_group" not provided (group to assign a positive score).')
@@ -73,9 +77,9 @@ def check_train(df: pd.DataFrame,
       return logging.error('At least one of the groups to classify is too small (n<10).')
     elif np.min(df[to_predict].value_counts()) < 100:
       logging.warn('At least one of the groups to classify may be too small (n<100).')
-    if np.sum((df[col_id]+df[to_predict]).duplicated()) > 0:
+    if np.sum((df['ID'].astype(str)+df[to_predict].astype(str)).duplicated()) > 0:
       logging.warn('Training dataset has duplicate participants.')
-    spare_type = 'classification'
+    mdl_type = 'SVM Classification'
 
   elif len(df[to_predict].unique()) > 2:
     if df[to_predict].dtype not in ['int64', 'float64']:
@@ -84,16 +88,16 @@ def check_train(df: pd.DataFrame,
       return logging.error('Sample size is too small (n<10).')
     elif len(df.index) < 100:
       logging.warn('Sample size may be too small (n<100).')
-    if np.sum(df[col_id].duplicated()) > 0:
+    if np.sum(df['ID'].duplicated()) > 0:
       logging.warn('Training dataset has duplicate participants.')
     if pos_group != '':
       logging.info('SPARE regression does not need a "pos_group". This will be ignored.')
-    spare_type = 'regression'
+    mdl_type = 'SVM Regression'
   else:
     return logging.error('Variable to predict has no variance.')
     
-  logging.info(f'Dataframe checked for SPARE {spare_type} training.')
-  return df, predictors, spare_type
+  logging.info(f'Dataframe checked for SPARE training ({mdl_type}).')
+  return df, predictors, mdl_type
 
 def check_test(df: pd.DataFrame, 
                meta_data: dict):
@@ -103,19 +107,21 @@ def check_test(df: pd.DataFrame,
     df: a pandas dataframe containing testing data.
     meta_data: a dictionary containing training information on its paired SPARE model.
   """
-  col_id, col_age = col_names(df,['ID','Age'])
+  logging_basic_config()
+  if not {'ID','Age','Sex'}.issubset(set(df.columns)):
+    return logging.error('Please check required columns: ID, Age, Sex.')
   if not set(meta_data['predictors']).issubset(df.columns):
     cols_not_found = sorted(set(meta_data['predictors']) - set(df.columns))
     return logging.error(f'Not all predictors exist in the input dataframe: {cols_not_found}')
-    
-  if (np.min(df[col_age]) < meta_data['age_range'][0]) or (
-           np.max(df[col_age]) > meta_data['age_range'][1]):
+  
+  if (np.min(df['Age']) < np.min((meta_data['cv_results']['Age']))) or (
+           np.max(df['Age']) > np.max((meta_data['cv_results']['Age']))):
     logging.warn('Some participants fall outside the age range of the SPARE model.')
 
   if np.sum(np.sum(pd.isna(df[meta_data['predictors']]))) > 0:
     logging.warn('Some participants have invalid predictor variables.')
 
-  if np.any(df[col_id].isin(meta_data['cv_results'][col_id])):
+  if np.any(df['ID'].isin(meta_data['cv_results']['ID'])):
     logging.info('Some participants seem to have been in the model training.')
 
 def smart_unique(df1: pd.DataFrame,
@@ -135,28 +141,26 @@ def smart_unique(df1: pd.DataFrame,
   Returns:
     a trimmed pandas dataframe or a tuple of two dataframes with only one time point per ID.
   """
+  logging_basic_config(verbose)
   assert (isinstance(df2, pd.DataFrame) or (df2 is None)), (
     'Either provide a 2nd pandas dataframe for the 2nd argument or specify it with "to_predict"')
-  if verbose == 0:
-    logging.basicConfig(level=logging.WARNING, format='%(levelname)s: %(message)s', force=True)
-  col_id = col_names(df1, cols=['ID'])
   if df2 is None:
     if to_predict is None:
       return logging.error('Either provide a second dataframe or provide a column "to_predict"')
-    elif len(df1[to_predict].unique()) > 2:
-      if ~np.any(df1[col_id].duplicated()):
+    if len(df1[to_predict].unique()) < 2:
+      return logging.error('Variable to predict has no variance.')
+    if len(df1[to_predict].unique()) > 2:
+      if ~np.any(df1['ID'].duplicated()):
         logging.info('No duplicated IDs.')
       else:
         logging.info('Select unique time points for SPARE regression training.')
         df1[f'{to_predict}_from_mean'] = np.abs(df1[to_predict] - np.mean(df1[to_predict]))
-        df1 = df1[df1.groupby(col_id)[f'{to_predict}_from_mean'].transform(
-                          max) == df1[f'{to_predict}_from_mean']].reset_index(drop=True)
-        df1 = df1.drop(columns=f'{to_predict}_from_mean')
-        df1 = df1[~df1[col_id].duplicated()].reset_index(drop=True)
+        df1 = (df1[df1.groupby('ID')[f'{to_predict}_from_mean']
+                      .transform(max) == df1[f'{to_predict}_from_mean']]
+                      .drop(columns=f'{to_predict}_from_mean'))
+        df1 = df1[~df1['ID'].duplicated()].reset_index(drop=True)
       return df1
-    elif len(df1[to_predict].unique()) < 2:
-      return logging.error('Variable to predict has no variance.')
-    if ~np.any((df1[col_id].astype(str) + df1[to_predict].astype(str)).duplicated()):
+    if ~np.any(df1.groupby(['ID', to_predict]).size() > 1):
       logging.info('No duplicated IDs in either group.')
       return df1
     grps = list(df1[to_predict].unique())
@@ -165,30 +169,26 @@ def smart_unique(df1: pd.DataFrame,
   else:
     if to_predict is not None:
       logging.info('"to_predict" will be ignored.')
-    if (~np.any(df1[col_id].duplicated())) and (~np.any(df2[col_id].duplicated())):
+    if (~np.any(df1['ID'].duplicated())) and (~np.any(df2['ID'].duplicated())):
       logging.info('No duplicated IDs in either group.')
       return (df1, df2)
     no_df2 = False
 
   logging.info('Select unique time points for SPARE classification training.')
-  col_age = col_names(df1, cols=['Age'])
   swap = False
-  if stats.ttest_ind(df1[col_age], df2[col_age]).pvalue < 0.05:
-    if np.mean(df1[col_age]) < np.mean(df2[col_age]):
+  if stats.ttest_ind(df1['Age'], df2['Age']).pvalue < 0.05:
+    if np.mean(df1['Age']) < np.mean(df2['Age']):
         df1, df2, swap = df2.copy(), df1.copy(), True
-    df2 = df2.loc[df2[col_age] >= np.min(df1[col_age])].reset_index(drop=True)
-    df1 = df1[df1.groupby(col_id)[col_age].transform(min) == df1[col_age]].reset_index(drop=True)
-    df2 = df2[df2.groupby(col_id)[col_age].transform(max) == df2[col_age]].reset_index(drop=True)
+    df2 = df2.loc[df2['Age'] >= np.min(df1['Age'])].reset_index(drop=True)
+    df1 = df1[df1.groupby('ID')['Age'].transform(min) == df1['Age']].reset_index(drop=True)
+    df2 = df2[df2.groupby('ID')['Age'].transform(max) == df2['Age']].reset_index(drop=True)
   else:
     logging.info('Age difference not significant between two groups.')
-  df1 = df1[~df1[col_id].duplicated()].reset_index(drop=True)
-  df2 = df2[~df2[col_id].duplicated()].reset_index(drop=True)
+  df1 = df1[~df1['ID'].duplicated()].reset_index(drop=True)
+  df2 = df2[~df2['ID'].duplicated()].reset_index(drop=True)
   if swap:
     df1, df2 = df2.copy(), df1.copy()
-  if no_df2:
-    return pd.concat([df1, df2], ignore_index=True)
-  else:
-    return (df1, df2)
+  return pd.concat([df1, df2], ignore_index=True) if no_df2 else (df1, df2)
 
 def age_sex_match(df1: pd.DataFrame,
                   df2: pd.DataFrame = None,
@@ -203,6 +203,7 @@ def age_sex_match(df1: pd.DataFrame,
     df2: a pandas dataframe (optional) if df1 and df2 are two groups to classify.
     to_match: a binary variable of two groups. Must be one of the columns in df.
       Ignored if df2 is given.
+      If to_match is 'Sex', then only perform age matching.
     p_threshold: minimum p-value for matching.
     verbose: whether to output messages.
     age_out_percentage: percentage of the larger group to randomly select a participant to
@@ -213,13 +214,13 @@ def age_sex_match(df1: pd.DataFrame,
   Returns:
     a trimmed pandas dataframe or a tuple of two dataframes with age/sex matched groups.
   """
+  logging_basic_config(verbose)
   assert (isinstance(df2, pd.DataFrame) or (df2 is None)), (
-    'Either provide a 2nd pandas dataframe for the 2nd argument or specify it with "to_match"')
-  if verbose == 0:
-    logging.basicConfig(level=logging.WARNING, format='%(levelname)s: %(message)s', force=True)
+    'Either provide a 2nd pandas dataframe for the 2nd argument or specify the two groups with "to_match"')
+
   if df2 is None:
     if to_match is None:
-      return logging.error('Either provide a second dataframe or provide a column "to_match"')
+      return logging.error('Either provide a 2nd dataframe or provide a column "to_match"')
     if len(df1[to_match].unique()) != 2:
       return logging.error('Variable to match must be binary')
     grps = list(df1[to_match].unique())
@@ -227,22 +228,28 @@ def age_sex_match(df1: pd.DataFrame,
     no_df2 = True
   else:
     if to_match is not None:
-      logging.info('"to_match" will be ignored.')
+      logging.info('Two dataframes provided. "to_match" will be ignored.')
     no_df2 = False
 
   if (age_out_percentage <= 0) or (age_out_percentage >= 100):
     return logging.error('Age-out-percentage must be between 0 and 100')
- 
+  if (len(df1['Sex'].unique())==1) & (len(df2['Sex'].unique())==1):
+    logging.info('Performing age matching only.')
+    sex_match = False
+  else:
+    sex_match = True
+
   swap = 1
   random.seed(2022)
   n_orig = len(df1.index) + len(df2.index)
-  col_age, col_sex = col_names(df1, ['Age', 'Sex'])
-  s1, s2 = df1[col_sex].unique()
 
-  p_age = stats.ttest_ind(df1[col_age], df2[col_age]).pvalue
-  p_sex = stats.chi2_contingency([np.array(df1[col_sex].value_counts()), np.array(df2[col_sex].value_counts())])[1]
-  if verbose > 1:
-    print(f' Orig.: P_age: {np.round(p_age,2)}/ P_sex {np.round(p_sex,2)}')
+  p_age = stats.ttest_ind(df1['Age'], df2['Age']).pvalue
+  if sex_match:
+    s1, s2 = df1['Sex'].unique() 
+    p_sex = stats.chi2_contingency([np.array(df1['Sex'].value_counts()), np.array(df2['Sex'].value_counts())])[1]
+  else:
+    p_sex = 1
+  logging.debug(f' Original: P_age: {np.round(p_age,2)}/ P_sex: {np.round(p_sex,2)}')
 
   p_age_all, p_sex_all = np.array(p_age), np.array(p_sex)
   while np.min([p_age, p_sex]) < p_threshold:
@@ -250,39 +257,41 @@ def age_sex_match(df1: pd.DataFrame,
       df1, df2 = df2.copy(), df1.copy()
       swap *= -1
     if p_age < p_threshold:
-      if np.mean(df1[col_age]) < np.mean(df2[col_age]):
-        i_age = df1[col_age] < np.percentile(df1[col_age], age_out_percentage)
+      if np.mean(df1['Age']) < np.mean(df2['Age']):
+        i_age = df1['Age'] < np.percentile(df1['Age'], age_out_percentage)
       else:
-        i_age = df1[col_age] > np.percentile(df1[col_age], 100-age_out_percentage)
+        i_age = df1['Age'] > np.percentile(df1['Age'], 100-age_out_percentage)
     else:
-      i_age = df1[col_age] >= 0
+      i_age = df1['Age'] >= 0
     if p_sex < p_threshold:
-      if np.sum(df1[col_sex] == s1)/np.sum(df1[col_sex] == s2) > np.sum(df2[col_sex] == s1)/np.sum(df2[col_sex] == s2):
-        i_sex = df1[col_sex] == s1
+      if np.sum(df1['Sex'] == s1)/np.sum(df1['Sex'] == s2) > np.sum(df2['Sex'] == s1)/np.sum(df2['Sex'] == s2):
+        i_sex = df1['Sex'] == s1
       else:
-        i_sex = df1[col_sex] == s2
+        i_sex = df1['Sex'] == s2
     else:
-      i_sex = df1[col_sex].isin([s1, s2])
+      i_sex = df1['Sex'].isin([s1, s2])
 
     try:
       df1 = df1.drop(random.sample(list(df1[i_age & i_sex].index), 1)).reset_index(drop=True)
     except:
-      if np.min([len(df1.index), len(df2.index)]) > 10:
-        print('Try increasing "age_out_percentage" parameter')
-      return logging.error('Matching failed...')
-    p_age = stats.ttest_ind(df1[col_age], df2[col_age]).pvalue
-    p_sex = stats.chi2_contingency([np.array(df1[col_sex].value_counts()), np.array(df2[col_sex].value_counts())])[1]
+      suggestion = 'Try increasing "age_out_percentage" parameter.' if np.min([len(df1.index), len(df2.index)]) > 10 else ''
+      return logging.error(f'Matching failed... {suggestion}')
+    p_age = stats.ttest_ind(df1['Age'], df2['Age']).pvalue
+    p_sex = stats.chi2_contingency([np.array(df1['Sex'].value_counts()), np.array(df2['Sex'].value_counts())])[1]
     p_age_all = np.append(p_age_all, p_age)
     p_sex_all = np.append(p_sex_all, p_sex)
   if swap == -1:
     df1, df2 = df2.copy(), df1.copy()
 
-  if verbose > 1:
-    n_dropped = n_orig - len(df1.index) - len(df2.index)
-    print(f' {n_dropped} participants excluded')
-    print(f' Final: P_age: {np.round(p_age,2)}/ P_sex {np.round(p_sex,2)}')
+  logging.debug(f' {n_orig - len(df1.index) - len(df2.index)} participants excluded')
+  logging.debug(f' Final: P_age: {np.round(p_age,2)}/ P_sex {np.round(p_sex,2)}')
   logging.info('Age/Sex matched!')
   if no_df2:
     return pd.concat([df1, df2], ignore_index=True)
   else:
-    return (df1, df2)
\ No newline at end of file
+    return (df1, df2)
+
+def logging_basic_config(verbose=1, content_only=False):
+  logging_level = {0:logging.WARNING, 1:logging.INFO, 2:logging.DEBUG}
+  fmt = ' %(message)s' if content_only else '%(levelname)s (%(funcName)s): %(message)s'
+  logging.basicConfig(level=logging_level[verbose], format=fmt, force=True)
\ No newline at end of file
